package joot.generation;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.jooq.util.AbstractDefinition;
import org.jooq.util.ColumnDefinition;
import org.jooq.util.Definition;
import org.jooq.util.ForeignKeyDefinition;
import org.jooq.util.GeneratorStrategy.Mode;
import org.jooq.util.JavaGenerator;
import org.jooq.util.JavaWriter;
import org.jooq.util.TableDefinition;
import org.jooq.util.UniqueKeyDefinition;

/**
 * Java generator adding properties to generated records,
 * to represent the table relationships induced by foreign key constraints.
 * <p>
 * Generated records will have a property (i.e a field and a pair of getter/setter)
 * for each foreign key constraint referencing the table primary key.
 * <p>
 * Association type will be one-to-one if the foreign key columns encompass a unique key,
 * and one-to-many otherwise.
 * <p>
 * Property type will be the record type of the foreign key table,
 * or a {@code List} thereof, depending of the association type.
 */
public class RelationAwareJavaGenerator extends JavaGenerator {
	
	@Override
	protected void generateRecordClassFooter(TableDefinition table, JavaWriter out) {
		
		List<RelationDefinition> relations = getRelations(table);
		if (relations.isEmpty()) return ;

		out.tab(1).header("Relationships");

		relations.forEach(p -> {
			printRelationProperty(out, p);
			out.println().tab(1).println("// --------------------------------");
		});

		printRelationsCascadeMethod(out, relations);
	}
	
	/**
	 * Returns a list of definitions for one-to-* relationships.
	 * Each relationship is defined by a FK referencing the PK of the given table.
	 */
	protected static List<RelationDefinition> getRelations(TableDefinition table) {
		UniqueKeyDefinition pk = table.getPrimaryKey();
		return pk == null
			? Collections.emptyList()
			: pk.getForeignKeys()
				.stream()
				.map(fk -> new RelationDefinition(
						fk,
						getRelationName(fk),
						isRelationUnique(fk)
					)
				).collect(Collectors.toList());
	}
	
	/**
	 * Returns {@code true} if a relationship based on the given FK is one-to-one (rather than one-to-many). 
	 */
	protected static boolean isRelationUnique(ForeignKeyDefinition fk) {
		return fk
				.getKeyTable()
				.getUniqueKeys()
				.stream()
				.anyMatch(uk -> fk.getKeyColumns().containsAll(uk.getKeyColumns()));
	}
	
	/**
	 * Returns the name of a relationship defined by the given FK.
	 */
	protected static String getRelationName(ForeignKeyDefinition fk) {
		return fk.getKeyTable().getName().replaceFirst("^" + Pattern.quote(fk.getReferencedTable().getName()) + "_?", "");
	}
	
//////// RELATIONS DEFINITION
	
	/**
	 * Prints relationship properties.
	 * Each property will have a field, a getter and a setter.
	 * One-to-many properties will be of type {@code List<targetRecordType>}.
	 */
	protected void printRelationProperty(JavaWriter out, RelationDefinition relation) {
		
		String javadoc = String.format("one-to-%s relationship defined by foreign key {@code %s}",
				relation.isUnique() ? "one" : "many",
				relation.getConstraint().getName()
		);
		
		String member = member(relation),
		       getter = getter(relation),
		       setter = setter(relation),
		       type   = type(relation.getTargetTable());
		
		out.println();
		
		if (relation.isUnique()) {
			out.tab(1).println("protected %s %s;", type, member);
		} else {
			// Not using type directly for first print,
			// since need to pass List class as print argument to add the import.
			// Would have been easier if JavaWriter.ref() method was public.
			out.tab(1).println("protected %s<%s> %s;", List.class, type, member);
			type = "List<" + type + ">";
		}

		out
			.tab(1).javadoc("Getter for %s", javadoc)
			.tab(1).println("public %s %s() { return %s; }",
					type,
					getter,
					member
			)

			.tab(1).javadoc("Setter for %s", javadoc)
			.tab(1).println("public %s %s(%s %s) { this.%s = %s; %s}",
					fluentSetters() ? type(relation.getParentTable()) : "void",
					setter,
					type,
					member,
					member,
					member,
					fluentSetters() ? "return this; " : ""
			);
	}
	
	
	
//////// RELATIONS UPDATE
	
	/**
	 * Prints a method to populate the PKs of all relationship targets with the parent PK.
	 */
	public void printRelationsCascadeMethod(JavaWriter out, List<RelationDefinition> relations) {
		
		out
			.tab(1).javadoc("Cascade this record PK to child records defined by one-to-one and one-to-many relationships.")
			.tab(1).println("public void cascadePrimaryKey() {");
		
		for (RelationDefinition relation : relations) {
		
			out.tab(2).println("{");
			
			if (relation.isUnique())
				out
					.tab(3).println("%s rec = %s();", type(relation.getTargetTable()), getter(relation))
					.tab(3).println("if (rec != null) {");
			else
				out
					.tab(3).println("%s<%s> recs = %s();", List.class, type(relation.getTargetTable()), getter(relation))
					.tab(3).println("if (recs != null) for (%s rec : recs) {", type(relation.getTargetTable()));
				
			Iterator<ColumnDefinition> parentIter = relation.getParentColumns().iterator();
			Iterator<ColumnDefinition> targetIter = relation.getTargetColumns().iterator();
			while (parentIter.hasNext() && targetIter.hasNext())
				out.tab(4).println("rec.%s(%s());", setter(targetIter.next()), getter(parentIter.next()));
			
			out.tab(3).println("}");
			out.tab(2).println("}");
		}
		
		out.tab(1).println("}");
	}
	
//////// MISC

	protected String member(Definition def) {
		return getStrategy().getJavaMemberName(def, Mode.RECORD);
	}

	protected String getter(Definition def) {
		return getStrategy().getJavaGetterName(def, Mode.RECORD);
	}

	protected String setter(Definition def) {
		return getStrategy().getJavaSetterName(def, Mode.RECORD);
	}

	protected String type(Definition def) {
		return getStrategy().getJavaClassName(def, Mode.RECORD);
	}

	/**
	 * A one-to-* relationship from a parent table to a target table,
	 * that is in the opposite direction of the underlying foreign key.
	 */
	public static class RelationDefinition extends AbstractDefinition {

		private final ForeignKeyDefinition fk;
		private final boolean unique;

		public RelationDefinition(ForeignKeyDefinition fk, String name, boolean unique) {
			super(fk.getReferencedTable().getDatabase(), fk.getReferencedTable().getSchema(), name);
			this.fk = fk;
			this.unique = unique;
		}
		
		public ForeignKeyDefinition getConstraint() {
			return fk;
		}
		
		public TableDefinition getParentTable() {
			return fk.getReferencedTable();
		}
		
		public TableDefinition getTargetTable() {
			return fk.getKeyTable();
		}

		public List<ColumnDefinition> getParentColumns() {
			return fk.getReferencedColumns();
		}
		
		public List<ColumnDefinition> getTargetColumns() {
			return fk.getKeyColumns();
		}

		public boolean isUnique() {
			return unique;
		}
		
		@Override
		public List<Definition> getDefinitionPath() {
			return Stream.concat(
					getParentTable().getDefinitionPath().stream(),
					Stream.of(this)
			).collect(Collectors.toList());
		}
		
	}
}
