package joot.tree;

import java.util.List;
import java.util.Optional;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Stream;

import org.jooq.Record;
import org.jooq.RecordType;
import org.jooq.Table;
import org.jooq.TableRecord;

/**
 * Link in a table tree (i.e a foreign key).
 *
 * @param <KR> Key type
 * @param <PR> Parent table type
 * @param <CR> Child table type
 */
// TODO Make the key a field rather than a super-class, so it's easier to understand
// TODO Add a TableNode somewhere, so trees can be deeper than 1
public abstract class TreeLink<KR extends Record, PR extends TableRecord<PR>, CR extends TableRecord<CR>> extends TableKey<CR, KR> {

	protected TreeLink(Table<CR> table, RecordType<KR> key) {
		super(table, key);
	}

	public static class OneToOne<KR extends Record, PR extends TableRecord<PR>, CR extends TableRecord<CR>> extends TreeLink<KR, PR, CR> {

		private final Function<PR, CR> getter;
		private final BiConsumer<PR, CR> setter;
		
		public OneToOne(Table<CR> table, RecordType<KR> key, Function<PR, CR> getter, BiConsumer<PR, CR> setter) {
			super(table, key);
			this.getter = getter;
			this.setter = setter;
		}
		
		public CR get(PR parent) {
			return getter.apply(parent);
		}
		
		public void set(PR parent, CR child) {
			setter.accept(parent, child);
		}
		
		public Optional<CR> optional(PR parent) {
			return Optional.ofNullable(get(parent));
		}

	}
	
	public static class OneToMany<KR extends Record, PR extends TableRecord<PR>, CR extends TableRecord<CR>> extends TreeLink<KR, PR, CR> {

		private final Function<PR, List<CR>> getter;
		private final BiConsumer<PR, List<CR>> setter;
		
		public OneToMany(Table<CR> table, RecordType<KR> key, Function<PR, List<CR>> getter, BiConsumer<PR, List<CR>> setter) {
			super(table, key);
			this.getter = getter;
			this.setter = setter;
		}
		
		public List<CR> get(PR parent) {
			return getter.apply(parent);
		}
		
		public void set(PR parent, List<CR> child) {
			setter.accept(parent, child);
		}
		
		public Stream<CR> stream(PR parent) {
			List<CR> children = get(parent);
			return children != null ? children.stream() : Stream.empty();
		}
		
		public void add(PR parent, CR child, Supplier<? extends List<CR>> supplier) {
			List<CR> children = get(parent);
			if (children == null) {
				children = supplier.get();
				set(parent, children);
			}
			children.add(child);
		}
		
	}
	
}
