package joot.tree;

import static joot.tree.util.JooqUtils.intoPrimaryKey;
import static org.jooq.impl.DSL.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Stream;

import org.jooq.Condition;
import org.jooq.DSLContext;
import org.jooq.Record;
import org.jooq.RecordType;
import org.jooq.Row;
import org.jooq.Table;
import org.jooq.TableRecord;

import joot.tree.TreeLink.OneToMany;
import joot.tree.TreeLink.OneToOne;
import joot.tree.action.DeleteAction;
import joot.tree.action.DumpAction;
import joot.tree.action.FetchAction;
import joot.tree.action.InsertAction;
import joot.tree.action.MergeAction;

/**
 * Node in a table tree (i.e a parent table).
 *
 * @param <KR> Key type
 * @param <PR> Parent table type
 */
// TODO Make the key a field rather than a super-class, so it's easier to understand.
public abstract class TreeNode<KR extends Record, PR extends TableRecord<PR>> extends TableKey<PR, KR> {

	private final List<TreeLink<KR, PR, ?>> links = new ArrayList<>();
	
	protected TreeNode(Table<PR> table, RecordType<KR> parentKey) {
		super(table, parentKey);
	}
	
//////// FOR TREE SUBCLASSES
	
	/**
	 * Add a relation to this node.
	 * To be used by subclasses at construction time.
	 */
	protected void link(TreeLink<KR, PR, ?> link) {
		if (!matches(link))
			throw new IllegalArgumentException("Parent and child keys don't match");
		links.add(link);
	}

//////// FOR ACTIONS
	
	/**
	 * Get all links from child tables.
	 */
	public Stream<TreeLink<KR, PR, ?>> getLinks() {
		return links.stream();
	}

	/**
	 * Get 1-1 links to child tables.
	 */
	public Stream<OneToOne<KR, PR, ?>> getOneToOneLinks() {
		return links.stream()
				.filter(OneToOne.class::isInstance)
				.map(OneToOne.class::cast);
	}
	
	/**
	 * Get 1-N links to child tables.
	 */
	public Stream<OneToMany<KR, PR, ?>> getOneToManyLinks() {
		return links.stream()
				.filter(OneToMany.class::isInstance)
				.map(OneToMany.class::cast);
	}
	
//////// FOR USER
	
	public Collection<PR> fetch(DSLContext dsl, Condition condition) {
		return new FetchAction<>(this, dsl).fetch(condition);
	}
	
	public Collection<PR> fetch(DSLContext dsl) {
		return fetch(dsl, trueCondition());
	}
	
	public PR fetchOne(DSLContext dsl, Condition condition) {
		return fetch(dsl, condition).iterator().next();
	}
	
	public PR fetchOne(DSLContext dsl, PR parent) {
		return fetchOne(dsl, condition(intoPrimaryKey(parent)));
	}
	
	public int insert(DSLContext dsl, PR parent) {
		return new InsertAction<>(this, dsl).execute(parent);
	}
	
	public int merge(DSLContext dsl, PR parent) {
		return new MergeAction<>(this, dsl).execute(parent);
	}
	
	public int delete(DSLContext dsl, Row key) {
		return new DeleteAction<>(this, dsl).execute(key);
	}
	
	public void dump() {
		new DumpAction<>(this).dump();
	}
	
	public void dump(PR parent) {
		new DumpAction<>(this).dump(parent);
	}
}
