package joot.tree.action;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Stream;

import org.jooq.DSLContext;
import org.jooq.Field;
import org.jooq.Record;
import org.jooq.RowN;
import org.jooq.Table;
import org.jooq.TableField;
import org.jooq.TableRecord;
import org.jooq.impl.DSL;

import joot.tree.TreeLink;
import joot.tree.TreeNode;
import joot.tree.util.JooqUtils;
import joot.tree.TreeLink.OneToMany;
import joot.tree.TreeLink.OneToOne;

public class MergeAction<KR extends Record, PR extends TableRecord<PR>> extends AbstractWriteAction<KR, PR> {

	public MergeAction(TreeNode<KR, PR> node, DSLContext dsl) {
		super(node, dsl);
	}

	@Override
	protected int parent() {
		return insertOrMerge(parent);
	}
	
	@Override
	protected <CR extends TableRecord<CR>> int child(OneToOne<KR, PR, CR> link) {
		CR child = link.get(parent);
		return child == null ? deleteChildren(link, null) : mergeChild(child, link);
	}
	
	@Override
	protected <CR extends TableRecord<CR>> int child(OneToMany<KR, PR, CR> link) {
		List<RowN> pks = new ArrayList<>();
		List<CR> children = link.get(parent);
		int count = 0;
		if (children != null) for (CR child : children) {
			count += mergeChild(child, link);
			pks.add(pkValuesRow(child));
		}
		count += deleteChildren(link, pks);
		return count;
	}
	
	/**
	 * Delete children of a parent table, except those with given PKs if any.
	 */
	private int deleteChildren(TreeLink<?, ?, ?> link, List<RowN> except) {
		if (except == null) except = Collections.emptyList();
		return dsl
				.deleteFrom(
						link.getTable()
				).where(
						link.fieldsRow().eq(node.valuesRow(parent)),
						pkFieldsRow(link.getTable()).notIn(except)
				).execute();
	}
	
	/**
	 * Copy the parent key into a child record and merge it.
	 */
	private <CR extends TableRecord<CR>> int mergeChild(CR child, TreeLink<KR, PR, CR> link) {
		node.copyValues(parent, child, link);
		return insertOrMerge(child);
	}

	/**
	 * Insert (if no PK) or merge (if PK) a record.
	 * Only changed fields will be taken into account.
	 */
	private <CR extends TableRecord<CR>> int insertOrMerge(CR rec) {
		// merge if PK present
		if (JooqUtils.isPrimaryKeyNotNull(rec)) {
			// PK fields
			List<TableField<CR, ?>> pk = rec.getTable().getPrimaryKey().getFields();
			// changed or PK fields
			Field<?>[] changed = Stream
					.of(rec.fields())
					.filter(f -> rec.changed(f) || pk.contains(f))
					.toArray(Field[]::new);
			// merging
			return dsl.mergeInto(rec.getTable()).columns(changed).values(rec.into(changed).intoArray()).execute();
		}
		// insert otherwise
		else {
			return JooqUtils.attach(rec, dsl).insert();
		}
	}
	
	private static RowN pkFieldsRow(Table<?> table) {
		return DSL.row(table.getPrimaryKey().getFieldsArray());
	}
	
	private static RowN pkValuesRow(TableRecord<?> rec) {
		return DSL.row(JooqUtils.intoPrimaryKey(rec).intoArray());
	}

}
