package joot.tree.node;

import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Stream;

import org.jooq.ForeignKey;
import org.jooq.Record;
import org.jooq.Table;
import org.jooq.TableRecord;
import org.jooq.impl.DSL;

import joot.tree.TreeNode;
import joot.tree.TreeLink.OneToMany;
import joot.tree.TreeLink.OneToOne;
import joot.tree.util.ReflectUtils;
import joot.tree.util.ReflectUtils.Getter;
import joot.tree.util.ReflectUtils.Setter;

/**
 * TreeNode implementation based on FKs and reflection.
 * <p>
 * Relations are deduced from FKs referencing this table's PK.
 * Relations are 1-1 if the FK is unique (i.e it contains a unique key), 1-N otherwise.
 * <p>
 * Getter/setters are found by reflection.
 * For each relation, first getter/setter pair with matching type will be used, whatever its name.
 * 
 * @param <PR> Parent table type
 * 
 * @see joot.generation.RelationAwareJavaGenerator
 */
public class ReflectNode<PR extends TableRecord<PR>> extends TreeNode<Record, PR> {

	public <CR extends TableRecord<CR>> ReflectNode(Table<PR> table) {
		this(table, table.getRecordType());
	}
	
	@SuppressWarnings("unchecked")
	public <CR extends TableRecord<CR>> ReflectNode(Table<PR> table, Class<? extends PR> recordClass) {
		
		super(
				table,
				DSL.recordType(table.getPrimaryKey().getFieldsArray())
		);
		
		table
			.getPrimaryKey()
			.getReferences()
			.stream()
			.filter(fk -> TableRecord.class.isAssignableFrom(fk.getTable().getRecordType()))
			.map(fk -> (ForeignKey<CR, PR>) fk)
			.forEach(fk -> createLink(recordClass, fk));
	}
	
//////// LINKS
	
	private <CR extends TableRecord<CR>> void createLink(Class<? extends PR> parent, ForeignKey<CR, PR> fk) {
	
		// association is unique if child side of the FK is unique
		boolean unique = fk
				.getTable()
				.getKeys()
				.stream()
				.anyMatch(uk -> fk.getFields().containsAll(uk.getFields()));
		
		// record type
		Class<? extends CR> child = fk.getTable().getRecordType();
		
		// one to one
		if (unique) {
			Accessor<PR, CR> acc = accessor(parent, child);
			if (acc != null) {
				link(new OneToOne<>(
						fk.getTable(),
						DSL.recordType(fk.getFieldsArray()),
						acc.getter,
						acc.setter
				));
			}
		}
		
		// one to many
		else {
			Accessor<PR, List<CR>> acc = accessor(parent, listClass(child), child);
			if (acc != null) {
				link(new OneToMany<>(
						fk.getTable(),
						DSL.recordType(fk.getFieldsArray()),
						acc.getter,
						acc.setter
				));
			}
		}
	}
	
	
	// Dirty way to get a hold on a Class<List<E>>
	@SuppressWarnings({ "unchecked", "rawtypes" })
	private static <E> Class<? extends List<E>> listClass(Class<? extends E> clazz) {
		return (Class<? extends List<E>>) (Class) List.class;
	}
	
//////// ACCESSOR
	
	private static class Accessor<T, U> {

		final Function<T, U> getter;
		final BiConsumer<T, U> setter;
		
		Accessor(Function<T, U> getter, BiConsumer<T, U> setter) {
			this.getter = getter;
			this.setter = setter;
		}
	}
	
	// TODO Improve that
	private static <T, U> Accessor<T, U> accessor(Class<? extends T> parent, Class<? extends U> child, Class<?>... childTypeArguments) {
		
		// any matching getter
		Getter<T, U> getter = Stream
				.of(parent.getMethods())
				.filter(m -> m.getName().startsWith("get"))
				.map(m -> ReflectUtils.<T, U>getter(m, parent, child, childTypeArguments))
				.filter(Objects::nonNull)
				.findAny()
				.orElse(null);
		if (getter == null) return null;
		
		// any matching setter for same property than getter
		String name = "set" + getter.getMethod().getName().substring(3);
		Setter<T, U> setter = Stream
				.of(parent.getMethods())
				.filter(m -> m.getName().equals(name))
				.map(m -> ReflectUtils.<T, U>setter(m, parent, child, childTypeArguments))
				.filter(Objects::nonNull)
				.findAny()
				.orElse(null);
		if (setter == null) return null;
		
		// done
		return new Accessor<>(getter, setter);
	}

}
