package joot.tree.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Stream;

public abstract class ReflectUtils {

//////// GETTER
	
	/**
	 * If {@code method} could be a getter for a property of type {@code rClass<rTypeArguments>},
	 * wraps it in a {@link Getter} instance. Otherwise returns {@code null}.
	 * <p>
	 * The method name is not taken into account ; only its modifiers and arguments/return types are.
	 * 
	 * @param method          A method to wrap in a getter
	 * @param tClass          The raw class of the target (the type defining the method) 
	 * @param rClass          The raw class of the property (the return type of the getter) 
	 * @param rTypeArguments  The type parameters if the property has a generic type
	 * 
	 * @return                A getter wrapping the given method, or {@code null};
	 */
	public static <T, R> Getter<T, R> getter(Method method, Class<? extends T> tClass, Class<? extends R> rClass, Class<?>... rTypeArguments) {
		
		if (Modifier.isStatic(method.getModifiers())) return null;
		if (method.getParameterTypes().length != 0)  return null;
		if (!method.getReturnType().equals(rClass)) return null;
		
		if (rTypeArguments.length > 0 && getRawTypeArguments(method.getGenericReturnType()).equals(rTypeArguments))
			return null;
		
		return new Getter<>(method, tClass, rClass);
	}
	
	/**
	 * Wrapper for a getter method.
	 *
	 * @param <T> Target type (the type defining the method)
	 * @param <R> Property type (the return type of the getter)
	 */
	public static class Getter<T, R> implements Function<T, R> {

		private final Method m;
		private final Class<? extends T> tc;
		private final Class<? extends R> rc;

		private Getter(Method m, Class<? extends T> tc, Class<? extends R> rc) {
			this.m = m;
			this.tc = tc;
			this.rc = rc;
		}

		@Override
		public R apply(T t) {
			try {
				return rc.cast(m.invoke(tc.cast(t)));
			} catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
				throw new RuntimeException(e);
			}
		}
		
		public Method getMethod() {
			return m;
		}
	}
	
//////// SETTER
	
	/**
	 * If {@code method} could be a setter for a property of type {@code uClass<uTypeArguments>},
	 * wraps it in a {@link Setter} instance. Otherwise returns {@code null}.
	 * <p>
	 * The method name is not taken into account ; only its modifiers and argument/return types are.
	 * 
	 * @param method          A method to wrap in a setter
	 * @param tClass          The raw class of the target (the type defining the method) 
	 * @param uClass          The raw class of the property (the return type of the setter argument) 
	 * @param uTypeArguments  The type parameters if the property has a generic type
	 * 
	 * @return                A setter wrapping the given method, or {@code null};
	 */
	public static <T, U> Setter<T, U> setter(Method method, Class<? extends T> tClass, Class<? extends U> uClass, Class<?>... uTypeArguments) {
		
		if (Modifier.isStatic(method.getModifiers())) return null;
		if (method.getParameterTypes().length != 1)  return null;
		if (!method.getParameterTypes()[0].equals(uClass)) return null;
		
		if (uTypeArguments.length > 0 && getRawTypeArguments(method.getGenericParameterTypes()[0]).equals(uTypeArguments))
			return null;
		
		return new Setter<>(method, tClass, uClass);
	}
	
	/**
	 * Wrapper for a setter method.
	 *
	 * @param <T> Target type (the type defining the method)
	 * @param <U> Property type (the type of the setter argument)
	 */
	public static class Setter<T, U> implements BiConsumer<T, U> {

		private final Method m;
		private final Class<? extends T> tc;
		private final Class<? extends U> uc;

		private Setter(Method m, Class<? extends T> tc, Class<? extends U> uc) {
			this.m = m;
			this.tc = tc;
			this.uc = uc;
		}

		@Override
		public void accept(T t, U u) {
			try {
				m.invoke(tc.cast(t), uc.cast(u));
			} catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
				throw new RuntimeException(e);
			}
		}
		
		public Method getMethod() {
			return m;
		}
	}
	
//////// MISC
	
	// TODO Handle more types than ParameterizedType
	private static Class<?>[] getRawTypeArguments(Type type) {
		if (type instanceof ParameterizedType)
			return Stream.of(((ParameterizedType) type).getActualTypeArguments()).map(ReflectUtils::getRawType).toArray(Class[]::new);
		else
			return new Class<?>[0];
	}
	
	// TODO Handle more types than ParameterizedType
	private static Class<?> getRawType(Type type) {
		if (type instanceof ParameterizedType)
			type = ((ParameterizedType) type).getRawType();
		return Class.class.cast(type);
	}

}
